/*
 * Open connection for network block device
 *
 * Copyright 1997,1998 Pavel Machek, distribute under GPL
 *  <pavel@atrey.karlin.mff.cuni.cz>
 *
 * Version 1.0 - 64bit issues should be fixed, now
 */

#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netinet/in.h>         /* sockaddr_in, htons, in_addr */
#include <netdb.h>              /* hostent, gethostby*, getservby* */
#include <stdio.h>
#include <fcntl.h>
#include <linux/nbd.h>
#include <syslog.h>

#define MY_NAME "nbd_client"
#include "cliserv.h"

int
opennet( char *name, int port )
{
int sock;
struct sockaddr_in xaddrin;
int xaddrinlen = sizeof( xaddrin );
struct hostent *hostn;
int size = 8;

hostn = gethostbyname( name );
if (!hostn) err( "Gethostname failed: %m\n" ); 

if ( (sock = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP))<0 ) err( "Socket failed: %m\n");

xaddrin.sin_family = AF_INET;
xaddrin.sin_port = htons( port );
xaddrin.sin_addr.s_addr = * ((int *) hostn->h_addr);
if ( (connect( sock, (struct sockaddr *) &xaddrin, xaddrinlen) < 0 )) err( "Connect: %m\n" ); 

setmysockopt( sock );
return sock;
}

int
main( int argc, char *argv[] )
{
int port, sock, nbd, size, one = 1;
__u64 magic, size64;
char buf[256] = "\0\0\0\0\0\0\0\0\0";
int swap = (argc>4);

logging();

if (argc<4) { fprintf( stderr, "Usage: host port nbd_device -swap\n" ); return 1; }
port = atoi( argv[2] ); 

sock = opennet( argv[1], port );

nbd = open( argv[3], O_RDWR );
if (nbd<0) err( "Can not open NBD: %m\n" ); 

printf( "Negotiation: " ); 
if (read( sock, buf, 8 )<0) err( "Failed/1: %m\n" ); 
if (strcmp( buf, INIT_PASSWD )) err( "INIT_PASSWD bad\n" );
printf( "." );
if (read( sock, &magic, sizeof( magic )) <0 ) err( "Failed/2: %m\n" );
magic = __be64_to_cpu( magic );
if (magic != cliserv_magic) err( "Not enough cliserv_magic\n" );
printf( "." );

if (read( sock, &size64, sizeof( size64 )) <0 ) err( "Failed/3: %m\n" );
size64 = __be64_to_cpu(size64);
if (size64 > (~0UL)) err( "Exported device is too big for me. Get 64-bit machine :-(\n" );
size = size64;
printf( "size = %d", size );
if (read( sock, &buf, 128 )<0) err( "Failed/4: %m\n" );
printf( "\n" );

if (ioctl( nbd, NBD_SET_SIZE, size ) <0) err( "Ioctl/1 failed: %m\n" ); 
ioctl( nbd, NBD_CLEAR_SOCK );
if (ioctl( nbd, NBD_SET_SOCK, sock ) <0) err( "Ioctl/2 failed: %m\n" );

#ifndef SO_SWAPPING
if (swap)
  err( "You have to compile me on machine with swapping patch enabled in order to use it later.\n" );
#else
if (swap)
  if (setsockopt( sock, SOL_SOCKET, SO_SWAPPING, &one, sizeof( int ))<0) err( "Could not enable swapping: %m");
#endif

if (ioctl( nbd, NBD_DO_IT )<0) fprintf( stderr, "Kernel call returned: %m\n" );
                          else fprintf( stderr, "Kernel call returned.\n" );
printf( "Closing: que, " );
ioctl( nbd, NBD_CLEAR_QUE );
printf( "sock, " );
ioctl( nbd, NBD_CLEAR_SOCK );
printf( "done\n" );
}


