/*
 * Open network connection to specified remote machine
 *
 * Copyright 1997 Pavel Machek, <pavel@atrey.karlin.mff.cuni.cz>
 * Copyright 1998 Josh Parsons, <josh@coombs.anu.edu.au>
 *
 */

/* 
 * Things I've changed:
 * - code tidyup
 * - endianness problems
 *        --- Josh
 */

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>         /* sockaddr_in, htons, in_addr */
#include <netdb.h>              /* hostent, gethostby*, getservby* */
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <linux/nbd.h>

int
opennet( char *name, int port ) {
   int sock;
   struct sockaddr_in xaddrin;
   int xaddrinlen = sizeof( xaddrin );
   struct hostent *hostn;
   int size = 8;

   hostn = gethostbyname( name );
   if (!hostn) { 
      fprintf( stderr, "Gethostname for %s failed: %m\n", name ); 
      return 0; 
   }

   if ( (sock = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP))<0 )
      { 
	 fprintf( stderr, "Socket failed: %m\n"); 
	 return 0; 
      }

   setsockopt( sock, SOL_SOCKET, SO_SNDBUF, &size, sizeof( int ));
   size =1;
   if (setsockopt( sock, SOL_TCP, TCP_NODELAY, &size, sizeof( int ))<0) 
      fprintf( stderr, "[TCP_NODELAY: %m]");
   size =1024;
   if (setsockopt( sock, SOL_TCP, TCP_MAXSEG, &size, sizeof( int ))<0) 
      fprintf( stderr, "[TCP_MAXSEG: %m]");
   xaddrin.sin_family = AF_INET;
   xaddrin.sin_port = htons( port );
   xaddrin.sin_addr.s_addr = * ((int *) hostn->h_addr);
   if ( (connect( sock, (struct sockaddr *) &xaddrin, xaddrinlen) < 0 )) { 
      fprintf( stderr, "Connect: %m\n" ); 
      return 0; 
   }
   return sock;
}

int
main( int argc, char *argv[] ) {
   int port, sock, nbd, size, magic, one = 1;
   char buf[9];
   int swap = (argc>4);
   
   if (argc<4) { 
      fprintf( stderr, "Usage: host port nbd_device [-swap]\n" ); 
      return 1; 
   }
   port = atoi( argv[2] ); 
   
   sock = opennet( argv[1], port );
   if (!sock) { 
      fprintf( stderr, "Can not connect to server.\n" ); 
      return 1; 
   }

   nbd = open( argv[3], O_RDWR );
   if (nbd<0) { 
      fprintf( stderr, "Can not open NBD: %m\n" ); 
      return 1; 
   }

   if (read( sock, buf, 8 )<0) { 
      fprintf( stderr, "Failed: %m\n" ); 
      return 1; 
   }
   if (read( sock, &magic, sizeof( magic )) <0 ) { 
      fprintf( stderr, "Failed: %m\n" ); 
      return 1; 
   }
   magic=ntohl(magic);
   if (magic != 0x81861253) {
      fprintf( stderr, "Not enough magic\n" );
      if(magic==0x53128681) 
	 fprintf( stderr, 
		  "Perhaps the other end is running a broken server?\n" );
      return 1; 
   }

   if (read( sock, &size, sizeof( size )) <0 ) { 
      fprintf( stderr, "Failed: %m\n" ); 
      exit(1); 
   }
   size=ntohl(size);
   if (ioctl( nbd, NBD_SET_SIZE, size ) <0 ) { 
      fprintf( stderr, "Ioctl failed: %m\n" ); 
      exit(1); 
   }
   if (ioctl( nbd, NBD_SET_SOCK, sock )<0) { 
      fprintf( stderr, "Error ioctl-ing: %m\n" ); 
      exit(1); 
   }

#ifdef SO_SWAPPING
   if (swap)
      if (setsockopt( sock, SOL_SOCKET, SO_SWAPPING, &one, sizeof( int ))<0) 
	 fprintf( stderr, "Could not enable swapping: %m");
#endif
   
   printf("nbdclient: device size = %d\n",size);

   if (ioctl( nbd, NBD_DO_IT )<0) 
      fprintf( stderr, "Kernel call returned: %m\n" );
   else fprintf( stderr, "Kernel call returned.\n" );

   ioctl( nbd, NBD_CLEAR_QUE );
   ioctl( nbd, NBD_CLEAR_SOCK );
   return 0;
}


