/*
 * Network Block Device - server
 *
 * Copyright 1996-1998 Pavel Machek, distribute under GPL
 *  <pavel@atrey.karlin.mff.cuni.cz>
 *
 * Version 1.0 - hopefully 64-bit-clean
 */
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netinet/in.h>         /* sockaddr_in, htons, in_addr */
#include <netdb.h>              /* hostent, gethostby*, getservby* */
#include <syslog.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <linux/nbd.h>

#define MY_NAME "nbd_server"
#include "cliserv.h"

#undef DODBG
//#define DEBUG( a ) printf( a )
#define DEBUG( a ) 

void inline
readit( int f, void *buf, int len )
{
int res;
while( len>0 )
  {
  DEBUG( "*" );
  if ((res = read( f, buf, len )) <= 0) err( "Read failed: %m" );
  len -= res;
  buf += res;
  }
}

void inline
writeit( int f, void *buf, int len )
{
int res;
while( len>0 )
  {
  DEBUG( "+" );
  if ((res = write( f, buf, len )) <= 0) err( "Write failed: %m" );
  len -= res;
  buf += res;
  }
}

int port;	/* Port I'm listening at */
int export;	/* File I'm exporting */
__u64 exportsize;	/* ...and its length */

void
cmdline( int argc, char *argv[] )
{
off_t es = 0;

if ((argc != 3) && (argc != 4)) { printf( "Usage: port file_to_export [size]\n" ); exit( 0 ); }
port = atoi( argv[1] ); 
if (argc == 4) es = atoi( argv[3] );

if ((export = open( argv[2], O_RDWR )) == -1) err( "Could not open exported file: %m" );
if (!es)
  if ((es = lseek(export, 0, SEEK_END)) == -1) err( "Could not find size of exported file: %m" );
printf( "Export size recognized as %d.\n", es );
exportsize = es;
}

int
connectme( void )
{
struct sockaddr_in addrin;
int addrinlen = sizeof( addrin );
int net, sock;

if ( (sock = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP))<0 ) err( "socket: %m"); 

DEBUG( "Waiting for connections... bind, " );
addrin.sin_family = AF_INET;
addrin.sin_port = htons( port );
addrin.sin_addr.s_addr = 0;
if (bind( sock, (struct sockaddr *) &addrin, addrinlen )<0) err( "bind: %m"); 
DEBUG( "listen, " );
if (listen( sock, 1 )<0) err( "listen: %m"); 
DEBUG( "accept, " );
if ((net = accept( sock, (struct sockaddr *) &addrin, &addrinlen ))<0) err( "accept: %m" ); 

return net;
}

#define SEND writeit( net, &reply, sizeof( reply ));
#define ERROR { reply.error = htonl(-1); SEND; reply.error = 0; lastpoint = -1; }
#define MAYBESEEK(a) \
	if (lastpoint != a) { \
		if (llseek( export, a, SEEK_SET )<0) err( "Can not seek locally!\n" ); \
		lastpoint = a; \
	} else { DEBUG( "@" ); }

int 
main(int argc, char *argv[])
{
int net;
long long lastpoint = -1;
struct nbd_request request;
struct nbd_reply reply;
char zeros[ 300 ];
int i;

bzero( zeros, 290 );
logging();
cmdline(argc, argv);

net = connectme();
setmysockopt( net );

if (write( net, INIT_PASSWD, 8 )<0) err( "Negotiation failed: %m\n" ); 
cliserv_magic = __cpu_to_be64( cliserv_magic );
if (write( net, &cliserv_magic, sizeof( cliserv_magic ) )<0) err( "Negotiation failed: %m\n" ); 
exportsize = __cpu_to_be64( exportsize );
if (write( net, &exportsize, sizeof( exportsize ) )<0) err( "Negotiation failed: %m\n" ); 
if (write( net, zeros, 128 )<0) err( "Negotiation failed: %m\n" ); 

DEBUG( "Entering request loop!\n" );
reply.magic = htonl( NBD_REPLY_MAGIC );
reply.error = 0;
while (1)
  {
  char buf[ 20480 ];
  int len;

#ifdef DODBG
  i++;
  printf( "%d: ", i );
#endif

  readit( net, &request, sizeof( request ));
  request.from = __be64_to_cpu( request.from );
  request.type = ntohl( request.type );
  len = ntohl( request.len );

  if (request.magic != htonl(NBD_REQUEST_MAGIC)) err( "Not enough magic.\n" );
  if (len>10240) err( "Request too big!\n" );
#ifdef DODBG
  printf( "%s from %d (%d) len %d, ", (request.type ? "WRITE" : "READ"), 
	  (int) request.from, (int) request.from / 512, len );
#endif
  memcpy( reply.handle, request.handle, sizeof( reply.handle ));
  if ((request.from<0) || ((request.from + len)>exportsize)) { 
    	DEBUG( "[RANGE!]" ); 
	ERROR;
	continue;
  }
  if (request.type)	/* WRITE */
	{
	DEBUG( "wr: net->buf, " );
	readit( net, buf, len );
	MAYBESEEK( request.from );
	DEBUG( "buf->exp, " );
	writeit( export, buf, len );
	lastpoint += len;
	SEND;
	continue;
	}

  /* READ */

  MAYBESEEK( request.from );
  DEBUG( "exp->buf, " );
  if (read( export, buf+sizeof( struct nbd_reply ), len )!=len) { 
  	DEBUG( "Read failed: %m" ); 
	ERROR;
  	continue;
  }
  lastpoint += len;

  DEBUG( "buf->net, " );
  memcpy( buf, &reply, sizeof( struct nbd_reply ));
  writeit( net, buf, len+sizeof( struct nbd_reply ));
  DEBUG( "OK!\n" );
  }
}
